// Author(s)     :  Camille Wormser, Jane Tournois, Pierre Alliez, Laurent Saboret

#ifndef AABB_NODE_H_INCLUDED
#define AABB_NODE_H_INCLUDED

#include <CGAL/Simple_cartesian.h>
#include <CGAL/Cartesian_converter.h>
#include <CGAL/intersections.h>

#include <limits>
#include <vector>
#include <cassert>


// Modified version of CGAL::AABB_node to be independent of a polyhedron class.
template <class TriangleKernel,           ///< Triangles' kernel
          class TriangleIterator,         ///< Value type must be Triangle_3
          class TriangleIteratorIterator, ///< Value type must be TriangleIterator
          class Kernel = CGAL::Simple_cartesian<typename TriangleKernel::FT>
                                          ///< Local kernel for fast bounding box intersections
>                                          
class AABB_node
{
public:

  // Bounding box with (double) floating point arithmetic
  typedef CGAL::Iso_cuboid_3<Kernel> Bbox;
  
  // Local kernel object types
  typedef typename Kernel::FT FT;
  typedef typename Kernel::Point_3 Point;

  // Global kernel object types
  typedef typename TriangleKernel::Triangle_3 Triangle;
  typedef typename TriangleKernel::Point_3 Triangle_point;
  
private:

  // Converter of number types, points and vectors
  typedef CGAL::Cartesian_converter<TriangleKernel, Kernel> Converter; 
  
private:

  // bounding box
  Bbox m_bbox;

  // if leaf: data
  TriangleIterator m_primitive;

  // else: children data as iterator range (stored only in the tree)...
  TriangleIteratorIterator m_begin, m_end;
  // ...and children nodes.
  AABB_node *m_left_child;
  AABB_node *m_right_child;

public:

  // life cycle
  AABB_node(TriangleIteratorIterator a, TriangleIteratorIterator b)
  {
    m_left_child = m_right_child = NULL;
    expand(a, b);
  }

  ~AABB_node()
  {
    cleanup_recurse();
  }

  // access to primitive
  TriangleIterator primitive()  { return m_primitive; }

  // access to children
  AABB_node* left_child()  { return m_left_child; }
  AABB_node* right_child() { return m_right_child; }

  bool is_leaf()
  {
    return m_left_child == NULL && m_right_child == NULL;
  }

  // deletes the whole subtree rooted at this node (except this node, of course).
  void cleanup_recurse()
  {
    delete m_left_child;
    delete m_right_child;
    m_left_child = m_right_child = NULL;
  }

private:

  // compute bbox for iterator range of input primitives
  void compute_bbox()
  {
    assert(m_begin != m_end);
    
    FT xmin,xmax,ymin,ymax,zmin,zmax;
    std::vector<TriangleIterator>::iterator it = m_begin;

    // init with any point in container
    TriangleIterator triangle = *it;
    const Triangle_point& any_point = (*triangle)[0];
    xmin = xmax = any_point.x();
    ymin = ymax = any_point.y();
    zmin = zmax = any_point.z();

    // compute bbox
    for(; it != m_end; it++)
    {
      TriangleIterator triangle = *it;
      xyz_min(triangle,xmin,ymin,zmin);
      xyz_max(triangle,xmax,ymax,zmax);
    }
    
    // Enlarge it a little to avoid missing intersections
    // due to floating point computation errors.
    FT epsilon = FT(100)*std::numeric_limits<FT>::epsilon();
    Point p(xmin-epsilon, ymin-epsilon, zmin-epsilon);
    Point q(xmax+epsilon, ymax+epsilon, zmax+epsilon);
    m_bbox = Bbox(p,q);
  }

  void xyz_min(TriangleIterator triangle,
               FT& xmin, FT& ymin, FT& zmin)
  {
    Triangle_point a = (*triangle)[0];
    Triangle_point b = (*triangle)[1];
    Triangle_point c = (*triangle)[2];
    xmin = (std::min)(xmin,(std::min)(a.x(),(std::min)(b.x(),c.x())));
    ymin = (std::min)(ymin,(std::min)(a.y(),(std::min)(b.y(),c.y())));
    zmin = (std::min)(zmin,(std::min)(a.z(),(std::min)(b.z(),c.z())));
  }

  void xyz_max(TriangleIterator triangle,
               FT& xmax, FT& ymax, FT& zmax)
  {
    Triangle_point a = (*triangle)[0];
    Triangle_point b = (*triangle)[1];
    Triangle_point c = (*triangle)[2];
    xmax = (std::max)(xmax,(std::max)(a.x(),(std::max)(b.x(),c.x())));
    ymax = (std::max)(ymax,(std::max)(a.y(),(std::max)(b.y(),c.y())));
    zmax = (std::max)(zmax,(std::max)(a.z(),(std::max)(b.z(),c.z())));
  }

  // builds the tree by recursive expansion.
  // [a,b[ is the range of primitives to be added to the tree.
  void expand(TriangleIteratorIterator a, TriangleIteratorIterator b)
  {
    m_begin = a;
    m_end = b;
    int range = std::distance(m_begin, m_end);

    compute_bbox();

    // sort primitives along longest axis aabb
    sort_primitives();

    if(range > 1)
    {
      // split
      m_left_child  = new AABB_node(m_begin,         m_begin+range/2);
      m_right_child = new AABB_node(m_begin+range/2, m_end);
    }
    else // this is a leaf node, set one primitive
    {
      m_primitive = *m_begin;
    }
  }

  static Triangle_point centroid(TriangleIterator triangle)
  {
    Triangle_point a = (*triangle)[0];
    Triangle_point b = (*triangle)[1];
    Triangle_point c = (*triangle)[2];
    return CGAL::centroid(a,b,c);
  }

  static bool lower_x(const TriangleIterator& f1,
                      const TriangleIterator& f2)
  {
    return centroid(f1).x() < centroid(f2).x();
  }
  static bool lower_y(const TriangleIterator& f1,
                      const TriangleIterator& f2)
  {
    return centroid(f1).y() < centroid(f2).y();
  }
  static bool lower_z(const TriangleIterator& f1,
                      const TriangleIterator& f2)
  {
    return centroid(f1).z() < centroid(f2).z();
  }

  void sort_primitives()
  {
    switch(longest_axis())
    {
      case 0: // sort along x
        std::sort(m_begin,m_end, lower_x);
        break;
      case 1: // sort along y
        std::sort(m_begin,m_end, lower_y);
        break;
      default: // sort along z
        std::sort(m_begin,m_end, lower_z);
    }
  }

  // Return the node's longest axis as 0 (X), 1 (Y) or 2 (Z)
  int longest_axis()
  {
    FT max_size = (std::max)(xsize(),(std::max)(ysize(),zsize()));
    if(max_size == xsize())
      return 0; // axis along x
    if(max_size == ysize())
      return 1; // axis along y
    return 2; // axis along z
  }

  // size of bounding box along each axis
  FT xsize() { return m_bbox.xmax() - m_bbox.xmin(); }
  FT ysize() { return m_bbox.ymax() - m_bbox.ymin(); }
  FT zsize() { return m_bbox.zmax() - m_bbox.zmin(); }

public:

  // --------------------RAY/SEGMENT ORACLES----------------------//

  // Test against primitive if leaf, against box if not
  template<class SegmentType>     ///< tested with Segment_3 and Ray_3
  bool do_intersect(const SegmentType& seg)
  {
    Converter convert;

    if(!is_leaf()) 
    {
      // check against bbox
      return CGAL::do_intersect(convert(seg),m_bbox);
    }
    else 
    {
      // optimization: check first against triangle's bbox
      if (!CGAL::do_intersect(convert(seg),m_bbox))
          return false;

      // check against input primitive 
      return CGAL::do_intersect(Triangle(*m_primitive), seg);
    }
  }

}; // class AABB_node


#endif // AABB_NODE_H_INCLUDED

