// Author(s)     :  Camille Wormser, Jane Tournois, Pierre Alliez, Laurent Saboret

#ifndef AABB_TREE_H_INCLUDED
#define AABB_TREE_H_INCLUDED

#include "AABB_node.h"

#include <CGAL/Simple_cartesian.h>

#include <stack>


// Modified version of CGAL::AABB_tree to be independent of a polyhedron class.
template <class TriangleKernel,           ///< Triangles' kernel
          class TriangleIterator,         ///< Value type must be Triangle_3
          class Kernel = CGAL::Simple_cartesian<typename TriangleKernel::FT>
                                          ///< Local kernel for fast bounding box intersections
>                                          
class AABB_tree
{
private:

  // Nodes of the tree
  typedef AABB_node<TriangleKernel, 
                    TriangleIterator,  
                    typename std::vector<TriangleIterator>::iterator,
                    Kernel> Node;

public:

  // Local kernel object types
  typedef typename Kernel::FT FT;
  typedef typename Kernel::Point_3 Point;

  // Global kernel object types
  typedef typename TriangleKernel::Triangle_3 Triangle;
  typedef typename TriangleKernel::Point_3 Triangle_point;
  typedef std::pair<Triangle_point, TriangleIterator> Point_with_triangle;

private:

  // set of triangle pointers (global to the tree)
  std::vector<TriangleIterator> m_data;

  // root node
  Node* m_root;

public:

  // life cycle
  AABB_tree() : m_root(NULL) {}
  AABB_tree(TriangleIterator first, TriangleIterator beyond)
  {
    m_root = NULL;
    build_faces(first, beyond);
  }

  ~AABB_tree()
  {
    cleanup();
  }

  void cleanup()
  {
    m_data.clear();
    delete m_root; m_root = NULL;
  }

  // build tree
  bool build_faces(TriangleIterator first, TriangleIterator beyond)
  {
    cleanup();
    set_face_data(first, beyond);
    if(!empty())
    {
      m_root = new Node(m_data.begin(), m_data.end());
      return true;
    }
    return false;
  }

private:

  void set_face_data(TriangleIterator first, TriangleIterator beyond)
  {
    unsigned int nbf = std::distance(first, beyond);
    m_data.reserve(nbf);
    for (TriangleIterator triangle = first; triangle != beyond; triangle++)
      m_data.push_back(triangle);
  }

public:

  bool empty()
  {
    return m_data.size() < 2; // TODO: change this requirement to < 1
  }

  // --------------------RAY/SEGMENT ORACLES----------------------//

  template<class SegmentType>    ///< tested with Segment_3 and Ray_3
  bool first_intersection(const SegmentType& seg, Point_with_triangle& pwh) const
  {
    std::stack<Node*> nodes;
    nodes.push(m_root);
    while(!nodes.empty())
    {
      Node *pNode = nodes.top();
      nodes.pop();

      // Test against primitive if leaf, against box if not
      if(pNode->do_intersect(seg)) 
      {
        if(pNode->is_leaf())
        {
          // found intersecting triangle
          Triangle t(*pNode->primitive());
          CGAL::Object inter = CGAL::intersection(t.supporting_plane(),seg);
          Triangle_point result;
          bool ok = CGAL::assign(result, inter);
          assert(ok);
          pwh = Point_with_triangle(result, pNode->primitive());
          return true; 
        }
        else // recurse (test against bbox only)
        {
          nodes.push(pNode->left_child());
          nodes.push(pNode->right_child());
        }
      }
    }
    return false;
  }


  template<class SegmentType,    ///< tested with Segment_3 and Ray_3
           class OutputIterator> ///< value type must be TriangleIterator
  OutputIterator find_intersections(const SegmentType& seg, OutputIterator output) const
  {
    std::stack<Node*> nodes;
    nodes.push(m_root);
    while(!nodes.empty())
    {
      Node *pNode = nodes.top();
      nodes.pop();

      // Test against primitive if leaf, against box if not
      if(pNode->do_intersect(seg)) 
      {
        if(pNode->is_leaf())
        {
          *output = pNode->primitive();
          output++; 
        }
        else // recurse (test against bbox only)
        {
          nodes.push(pNode->left_child());
          nodes.push(pNode->right_child());
        }
      }
    }
    return output;
  }

  template<class SegmentType>     ///< tested with Segment_3 and Ray_3
  unsigned int nb_intersections(const SegmentType& seg) const
  {
    std::vector<TriangleIterator> intersections;
    find_intersections(seg, std::back_inserter(intersections));
    return intersections.size();
  }

}; // class AABB_tree


#endif // AABB_TREE_H_INCLUDED
